from arcgis.features import FeatureLayerCollection

url = 'https://services1.arcgis.com/bqfNVPUK3HOnCFmA/arcgis/rest/services/Police_Calls_for_Service/FeatureServer'
layer_collection = FeatureLayerCollection(url)

# pandas version for both cases is 1.4.3
# Fails with version 2.0.1
# Works with version 2.0.0
layer_collection.layers[0].query(where="DateTimeOfCall >= '2021-01-01' AND  DateTimeOfCall < '2021-12-31T23:59:59.999'",
    result_offset=2000, result_record_count=2000, return_all_records=False, as_df=True)


import os.path as path
import pandas as pd
from datetime import datetime
from packaging import version

if __name__ == '__main__':
    import data_loaders
    import _datasets
    # import preproc
    from defs import TableType, DataType, MULTI
    from _version import __version__
    import exceptions
else:
    from . import data_loaders
    from . import _datasets
    from . import __version__
    # from . import preproc
    from .defs import TableType, DataType, MULTI
    from . import exceptions

class Table:
    """
    A class that contains a DataFrame for a dataset along with meta information

    Attributes
    ----------
    details : pandas Series
        Series containing information about the dataset
    state : str
        Name of state where agencies in table are
    source_name : str
        Name of source
    agency : str
        Name of agency
    table_type : TableType enum
        Type of data contained in table
    year : int, list, MULTI
        Indicates years contained in table
    description : str
        Description of data source
    url : str
        URL where table was accessed from
    table : pandas of geopandasDataFrame
        Data accessed from source

    Methods
    -------
    to_csv(output_dir=None, filename=None)
        Convert table to CSV file
    get_csv_filename()
        Get default name of CSV file
    """

    details = None
    state = None
    source_name = None
    agency = None
    table_type = None
    year = None
    description = None
    url = None

    # Data
    table = None

    # From source
    _data_type = None
    _dataset_id = None
    _date_field = None
    _agency_field = None

    def __init__(self, source, table=None, year_filter=None, agency=None):
        '''Construct Table object
        This is intended to be generated by the Source.load_from_url and Source.load_from_csv classes

        Parameters
        ----------
        source : pandas or geopandas Series
            Series containing information on the source
        table : pandas or geopandas 
            Name of state where agencies in table are
        '''
        if not isinstance(source, pd.core.frame.DataFrame) and \
            not isinstance(source, pd.core.series.Series):
            raise TypeError("data must be an ID, DataFrame or Series")
        elif isinstance(source, pd.core.frame.DataFrame):
            if len(source) == 0:
                raise LookupError("DataFrame is empty")
            elif len(source) > 1:
                raise LookupError("DataFrame has more than 1 row")

            source = source.iloc[0]

        self.details = source
        self.table = table

        self.state = source["State"]
        self.source_name = source["SourceName"]

        if agency != None:
            self.agency = agency
        else:
            self.agency = source["Agency"]

        self.table_type = TableType(source["TableType"])  # Convert to Enum

        if year_filter != None:
            self.year = year_filter
        else:
            self.year = source["Year"]

        self.description = source["Description"]
        self.url = source["URL"]
        self._data_type = DataType(source["DataType"])  # Convert to Enum

        if not pd.isnull(source["dataset_id"]):
            self._dataset_id = source["dataset_id"]

        if not pd.isnull(source["date_field"]):
            self._date_field = source["date_field"]
        
        if not pd.isnull(source["agency_field"]):
            self._agency_field = source["agency_field"]


    def __repr__(self) -> str:
        skip = ["details", "table"]
        return ',\n'.join("%s: %s" % item for item in vars(self).items() if (item[0] not in skip and item[0][0] != "_"))

    def to_csv(self, output_dir=None, filename=None):
        '''Export table to CSV file. Use default filename for data that will
        be reloaded as an openpolicedata Table object

        Parameters
        ----------
        output_dir - str
            (Optional) Output directory. Default: current directory
        filename - str
            (Optional) Filename. Default: Result of get_csv_filename()
        '''
        if filename == None:
            filename = self.get_csv_filename()
        if output_dir != None:
            filename = path.join(output_dir, filename)
        if not isinstance(self.table, pd.core.frame.DataFrame):
            raise ValueError("There is no table to save to CSV")

        self.table.to_csv(filename, index=False, errors="surrogateescape")

        return filename


    def get_csv_filename(self):
        '''Generate default filename based on table parameters

        Returns
        -------
        str
            Filename
        '''
        return get_csv_filename(self.state, self.source_name, self.agency, self.table_type, self.year)


class Source:
    """
    Class for exploring a data source and loading its data

    ...

    Attributes
    ----------
    datasets : pandas or geopandas DataFrame
        Contains information on datasets available from the source

    Methods
    -------
    get_tables_types()
        Get types of data availble from the source
    get_years()
        Get years available for 1 or more datasets
    get_agencies()
        Get agencies available for 1 or more datasets
    load_from_url()
        Load data from URL
    load_from_csv()
        Load data from a previously saved CSV file
    """

    datasets = None
    __limit = None

    def __init__(self, source_name, state=None):
        '''Constructor for Source class

        Parameters
        ----------
        source_name - str
            Source name from datasets table
        state - str
            (Optional) Name of state. Only necessary if source_name is not unique.

        Returns
        -------
        Source object
        '''
        self.datasets = _datasets.datasets_query(source_name=source_name, state=state)

        # Ensure that all sources are from the same state
        if len(self.datasets) == 0:
            raise ValueError("No Sources Found")
        elif self.datasets["State"].nunique() > 1:
            raise ValueError("Not all sources are from the same state")


    def __repr__(self) -> str:
        return str(self.datasets)


    def get_tables_types(self):
        '''Get types of data availble from the source

        Returns
        -------
        list
            List containing types of data available from source
        '''
        return list(self.datasets["TableType"].unique())


    def get_years(self, table_type):
        '''Get years available for 1 or more datasets

        Parameters
        ----------
        table_type - str or TableType enum
            Only returns years for requested table type

        Returns
        -------
        list
            List of years available for 1 or more datasets
        '''
        if isinstance(table_type, TableType):
            table_type = table_type.value

        dfs = self.datasets
        if table_type != None:
            dfs = self.datasets[self.datasets["TableType"]==table_type]

        all_years = list(dfs["Year"])
        years = set()
        for k in range(len(all_years)):
            if all_years[k] != MULTI:
                years.add(all_years[k])
            else:
                df = dfs.iloc[k]
                _check_version(df)
                data_type =DataType(df["DataType"])
                url = df["URL"]
                if not pd.isnull(df["date_field"]):
                    date_field = df["date_field"]
                else:
                    raise ValueError("No date_field is provided to identify the years")
                
                if data_type ==DataType.CSV:
                    raise NotImplementedError("This needs to be tested before use")
                    if force_read:                    
                        table = pd.read_csv(url, parse_dates=True)
                        new_years = table[date_field].dt.year
                        new_years = new_years.unique()
                    else:
                        raise ValueError("Getting the year of a CSV files requires reading in the whole file. " +
                                        "Loading in the table may be a better option. If getYears is still desired " +
                                        " for this case, use forceRead=True")    
                elif data_type ==DataType.ArcGIS:
                        new_years = data_loaders.get_years_argis(url, date_field)
                elif data_type ==DataType.SOCRATA:
                        new_years = data_loaders.get_years_socrata(url, df["dataset_id"], date_field)
                else:
                    raise ValueError(f"Unknown data type: {data_type}")

                years.update(new_years)
            
        years = list(years)
        years.sort()

        return years


    def get_agencies(self, table_type=None, year=None, partial_name=None):
        '''Get agencies available for 1 or more datasets

        Parameters
        ----------
        table_type - str or TableType enum
            (Optional) If set, only returns agencies for requested table type
        year - int or the string "MULTI" or "N/A"
            (Optional)  If set, only returns agencies for requested year
        table_type - str or TableType enum
            (Optional) If set, only returns agencies for requested table type
        partial_name - str
            (Optional)  If set, only returns agencies containing the substring
            partial_name for datasets that contain multiple agencies

        Returns
        -------
        list
            List of agencies available for 1 or more datasets
        '''

        if isinstance(table_type, TableType):
            table_type = table_type.value

        src = self.datasets
        if table_type != None:
            src = self.datasets[self.datasets["TableType"]==table_type]

        if year != None:
            src = src[src["Year"] == year]

        if len(src) == 1:
            src = src.iloc[0]
        else:
            raise ValueError("table_type and year inputs must filter for a single source")            

        # If year is multi, need to use self._agencyField to query URL
        # Otherwise return self.agency
        if src["Agency"] == MULTI:
            _check_version(src)
            data_type =DataType(src["DataType"])
            if data_type ==DataType.CSV:
                raise NotImplementedError(f"Unable to get agencies for {data_type}")
            elif data_type ==DataType.ArcGIS:
                raise NotImplementedError(f"Unable to get agencies for {data_type}")
            elif data_type ==DataType.SOCRATA:
                if partial_name is not None:
                    opt_filter = src["agency_field"] + " LIKE '%" + partial_name + "%'"
                else:
                    opt_filter = None

                select = "DISTINCT " + src["agency_field"]
                if year == MULTI:
                    year = None
                agencySet = data_loaders.load_socrata(src["URL"], src["dataset_id"], 
                    date_field=src["date_field"], year=year, opt_filter=opt_filter, select=select, output_type="set")
                return list(agencySet)
            else:
                raise ValueError(f"Unknown data type: {data_type}")
        else:
            return [src["Agency"]]
        

    def load_from_url(self, year, table_type=None, agency=None, pbar=True):
        '''Load data from URL

        Parameters
        ----------
        year - int or length 2 list or the string "MULTI" or "N/A"
            Used to identify the requested dataset if equal to its year value
            Otherwise, for datasets containing multiple years, this filters 
            the return data for a specific year (int input) or a range of years
            [X,Y] to return data for years X to Y
        table_type - str or TableType enum
            (Optional) If set, requested dataset will be of this type
        agency - str
            (Optional) If set, for datasets containing multiple agencies, data will
            only be returned for this agency
        pbar - bool
            (Optional) Whether to show progress bar when loading data. Default True

        Returns
        -------
        Table
            Table object containing the requested data
        '''

        return self.__load(year, table_type, agency, True, pbar)

    def __load(self, year, table_type, agency, load_table, pbar=True):
        if isinstance(table_type, TableType):
            table_type = table_type.value

        src = self.datasets
        if table_type != None:
            src = src[self.datasets["TableType"] == table_type]

        if isinstance(year, list):
            matchingYears = src["Year"] == year[0]
            for y in year[1:]:
                matchingYears = matchingYears | (src["Year"] == y)
        else:
            matchingYears = src["Year"] == year

        filter_by_year = not matchingYears.any()
        if not filter_by_year:
            # Use source for this specific year if available
            src = src[matchingYears]
        else:
            # If there are not any years corresponding to this year, check for a table
            # containing multiple years
            src = src.query("Year == '" + MULTI + "'")

        if isinstance(src, pd.core.frame.DataFrame):
            if len(src) == 0:
                raise ValueError(f"There are no sources matching tableType {table_type} and year {year}")
            elif len(src) > 1:
                raise ValueError(f"There is more than one source matching tableType {table_type} and year {year}")
            else:
                src = src.iloc[0]

        # Load data from URL. For year or agency equal to multi, filtering can be done
        data_type =DataType(src["DataType"])
        url = src["URL"]

        if filter_by_year:
            year_filter = year
        else:
            year_filter = None

        if not pd.isnull(src["dataset_id"]):
            dataset_id = src["dataset_id"]

        table_year = None
        if not pd.isnull(src["date_field"]):
            date_field = src["date_field"]
            if year_filter != None:
                table_year = year_filter
        else:
            date_field = None
        
        table_agency = None
        if not pd.isnull(src["agency_field"]):
            agency_field = src["agency_field"]
            if agency != None and data_type !=DataType.ArcGIS:
                table_agency = agency
        else:
            agency_field = None
        
        #It is assumed that each data loader method will return data with the proper data type so date type etc...
        if load_table:
            _check_version(src)
            if data_type ==DataType.CSV:
                table = data_loaders.load_csv(url, date_field=date_field, year_filter=year_filter, 
                    agency_field=agency_field, agency=agency, limit=self.__limit, pbar=pbar)
            elif data_type ==DataType.ArcGIS:
                table = data_loaders.load_arcgis(url, date_field, year_filter, limit=self.__limit, pbar=pbar)
            elif data_type ==DataType.SOCRATA:
                opt_filter = None
                if agency != None and agency_field != None:
                    # Double up any apostrophes for SQL query
                    agency = agency.replace("'","''")
                    opt_filter = agency_field + " = '" + agency + "'"

                table = data_loaders.load_socrata(url, dataset_id, date_field=date_field, year=year_filter, opt_filter=opt_filter, 
                    limit=self.__limit, pbar=pbar)
            else:
                raise ValueError(f"Unknown data type: {data_type}")

            table = _check_date(table, date_field)                        
        else:
            table = None

        return Table(src, table, year_filter=table_year, agency=table_agency)


    def load_from_csv(self, year, output_dir=None, table_type=None, agency=None):
        '''Load data from previously saved CSV file
        
        Parameters
        ----------
        year - int or length 2 list or the string "MULTI" or "N/A"
            Used to identify the requested dataset if equal to its year value
            Otherwise, for datasets containing multiple years, this filters 
            the return data for a specific year (int input) or a range of years
            [X,Y] to return data for years X to Y
        output_dir - str
            (output_dirOptional) Directory where CSV file is stored
        table_type - str or TableType enum
            (Optional) If set, requested dataset will be of this type
        agency - str
            (Optional) If set, for datasets containing multiple agencies, data will
            only be returned for this agency

        Returns
        -------
        Table
            Table object containing the requested data
        '''

        table = self.__load(year, table_type, agency, False)

        filename = table.get_csv_filename()
        if output_dir != None:
            filename = path.join(output_dir, filename)            

        table.table = pd.read_csv(filename, parse_dates=True)
        table.table = _check_date(table.table, table._date_field)  

        return table

    def get_csv_filename(self, year, output_dir=None, table_type=None, agency=None):
        '''Get auto-generated CSV filename
        
        Parameters
        ----------
        year - int or length 2 list or the string "MULTI" or "N/A"
            Used to identify the requested dataset if equal to its year value
            Otherwise, for datasets containing multiple years, this filters 
            the return data for a specific year (int input) or a range of years
            [X,Y] to return data for years X to Y
        output_dir - str
            (Optional) Directory where CSV file is stored
        table_type - str or TableType enum
            (Optional) If set, requested dataset will be of this type
        agency - str
            (Optional) If set, for datasets containing multiple agencies, data will
            only be returned for this agency

        Returns
        -------
        str
            Auto-generated CSV filename
        '''

        table = self.__load(year, table_type, agency, False)

        filename = table.get_csv_filename()
        if output_dir != None:
            filename = path.join(output_dir, filename)             

        return filename

def _check_date(table, date_field):
    if date_field != None and len(table)>0:
        dts = table[date_field]
        dts = dts[dts.notnull()]
        if len(dts) > 0:
            one_date = dts.iloc[0]            
            if type(one_date) == str:
                table = table.astype({date_field: 'datetime64[ns]'})
            elif date_field.lower() == "year":
                try:
                    float(one_date)
                except:
                    raise
                
                table[date_field] = table[date_field].apply(lambda x: datetime(x,1,1))
                
            # Replace bad dates with NaT
            table[date_field].replace(datetime.strptime('1900-01-01 00:00:00', '%Y-%m-%d %H:%M:%S'), pd.NaT, inplace=True)


    return table


def get_csv_filename(state, source_name, agency, table_type, year):
    '''Get default CSV filename for the given parameters. Enables reloading of data from CSV.
    
    Parameters
    ----------
    state - str
        Name of state
    source_name - str
        Name of source
    agency - str
        Name of agency
    table_type - str or TableType enum
        Type of data
    year = int or length 2 list or the string "MULTI" or "N/A"
        Year of data to load, range of years of data to load as a list [X,Y]
        to load years X to Y, or a string to indicate all of multiple year data
        ("MULTI") or a dataset that has no year filtering ("N/A")

    Returns
    -------
    str
        Default CSV filename
    '''
    if isinstance(table_type, TableType):
        table_type = table_type.value
        
    filename = f"{state}_{source_name}"
    if source_name != agency:
        filename += f"_{agency}"
    filename += f"_{table_type}"
    if isinstance(year, list):
        filename += f"_{year[0]}_{year[-1]}"
    else:
        filename += f"_{year}"

    # Clean up filename
    filename = filename.replace(",", "_").replace(" ", "_").replace("__", "_").replace("/", "_")

    filename += ".csv"

    return filename

def _check_version(df):
    min_version = df["min_version"] 
    if pd.notnull(min_version):
        src_name = df["SourceName"]
        state = df["State"]
        table_type = df["TableType"]
        year = df["Year"]
        if min_version == "-1":
            raise exceptions.OPD_FutureError(
                f"Year {year} {table_type} data for {src_name} in {state} cannot be loaded in this version. " + \
                    "It will be made available in a future release"
            )
        elif version.parse(min_version) < version.parse(__version__):
            raise exceptions.OPD_MinVersionError(
                f"Year {year} {table_type} data for {src_name} in {state} cannot be loaded in version {__version__} of openpolicedata. " + \
                    f"Update OpenPoliceData to at least version {min_version} to access this data."
            )

if __name__ == '__main__':
    istart = 182
    from datetime import date
    datasets = _datasets.datasets_query()
    max_num_stanford = 1
    num_stanford = 0
    prev_sources = []
    prev_tables = []
    output_dir = ".\\data"
    action = "standardize"
    issue_datasets = ["Austin", "Chapel Hill", "Fayetteville", "San Diego"]
    is_austin = datasets["SourceName"].apply(lambda x : x in issue_datasets)
    not_austin = datasets["SourceName"].apply(lambda x : x not in issue_datasets)
    # Austin has unknown race values. Emailed dataset owner.
    datasets = pd.concat([datasets[not_austin], datasets[is_austin]])
    for i in range(istart, len(datasets)):
        if "stanford.edu" in datasets.iloc[i]["URL"]:
            num_stanford += 1
            if num_stanford > max_num_stanford:
                continue

        srcName = datasets.iloc[i]["SourceName"]
        state = datasets.iloc[i]["State"]

        if datasets.iloc[i]["Agency"] == MULTI and srcName == "Virginia":
            # Reduce size of data load by filtering by agency
            agency = "Fairfax County Police Department"
        else:
            agency = None

        skip = False
        for k in range(len(prev_sources)):
            if srcName == prev_sources[k] and datasets.iloc[i]["TableType"] ==prev_tables[k]:
                skip = True

        if skip:
            continue

        prev_sources.append(srcName)
        prev_tables.append(datasets.iloc[i]["TableType"])

        table_print = datasets.iloc[i]["TableType"]
        now = datetime.now().strftime("%d.%b %Y %H:%M:%S")
        print(f"{now} Saving CSV for dataset {i} of {len(datasets)}: {srcName} {table_print} table")

        src = Source(srcName, state=state)

        if action == "standardize":
            if datasets.iloc[i]["DataType"] ==DataType.CSV.value:
                table = src.load_from_csv(datasets.iloc[i]["Year"], table_type=datasets.iloc[i]["TableType"])
            else:
                year = date.today().year
                table = None
                for y in range(year, year-20, -1):
                    try:
                        csv_filename = src.get_csv_filename(y, output_dir, datasets.iloc[i]["TableType"], 
                            agency=agency)
                    except ValueError as e:
                        if "There are no sources matching tableType" in e.args[0]:
                            continue
                        else:
                            raise
                    except:
                        raise
                    
                    if path.exists(csv_filename):
                        table = src.load_from_csv(y, table_type=datasets.iloc[i]["TableType"], 
                            agency=agency,output_dir=output_dir)
                        break

            table.standardize()
        else:
            if datasets.iloc[i]["DataType"] ==DataType.CSV.value:
                csv_filename = src.get_csv_filename(datasets.iloc[i]["Year"], output_dir, datasets.iloc[i]["TableType"])
                if path.exists(csv_filename):
                    continue
                table = src.load_from_url(datasets.iloc[i]["Year"], datasets.iloc[i]["TableType"])
            else:
                years = src.get_years(datasets.iloc[i]["TableType"])
                
                if len(years)>1:
                    # It is preferred to to not use first or last year that start and stop of year are correct
                    year = years[-2]
                else:
                    year = years[0]

                csv_filename = src.get_csv_filename(year, output_dir, datasets.iloc[i]["TableType"], 
                                        agency=agency)

                if path.exists(csv_filename):
                    continue

                table = src.load_from_url(year, datasets.iloc[i]["TableType"], 
                                        agency=agency)

            table.to_csv(".\\data")

    print("data main function complete")



# Definition of constants
from __future__ import annotations
from enum import Enum
from collections import namedtuple

# These are the types of data currently available in opd.
# They all have corresponding data loaders in data_loaders.py
# When new data loaders are added, this list should be updated.
class DataType(Enum):
    CSV = "CSV"
    ArcGIS = "ArcGIS"
    SOCRATA = "Socrata"

# These are the types of tables currently available in opd.
# Add to this list when datasets do not correspond to the below data types
class TableType(str, Enum):
    # Adding a description property in addition to value
    # https://rednafi.github.io/reflections/add-additional-attributes-to-enum-members-in-python.html
    description = None

    def __new__(
        cls, value: str, description: str = ""
    ) -> TableType:

        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.description = description
        return obj

    # Below tuples are (value, description)
    ARRESTS = ("ARRESTS", "Data describing arrests of individuals.")
    # Definition from https://www.policedatainitiative.org/datasets/calls-for-service/
    CALLS_FOR_SERVICE = ("CALLS FOR SERVICE", '"Calls for service to law enforcement agencies generally include calls ' +
        "to “911” for emergency assistance and may also include calls to non-emergency numbers. Generally, “311” data " +
        'is not included in calls for service data." https://www.policedatainitiative.org/datasets/calls-for-service/')
    # Definition from https://police.laws.com/police/police-citation
    CITATIONS = ("CITATIONS",'"A police citation, which is commonly called a "ticket", is official documentation stating ' +
        "that an individual has been made aware of a violation by an officer of the law. Police citations can be issued in " +
        "a variety of spectrums, ranging from vehicular violations to civil violations. This table is for citations resulting " +
        'from both traffic and pedestrian stops." https://police.laws.com/police/police-citation')
    # Defintion from https://www.policedatainitiative.org/datasets/complaints/
    COMPLAINTS = ("COMPLAINTS",'"An important aspect of accountable public service is citizen recourse and the ability ' +
        "for the community to share information regarding any potential mistreatment, including situations that may violate " +
        "agency policies and procedures and/or local ordinances or law. National law enforcement accreditation standards call " +
        "for agencies to have a process for collecting complaint information and have a process for making data about the " +
        'complaints available to the community." https://www.policedatainitiative.org/datasets/complaints/')
    # Definition from https://www.policedatainitiative.org/datasets/agency-workforce-demographics/
    EMPLOYEE = ("EMPLOYEE","Demographic data of the police workforce")
    FIELD_CONTACTS = ("FIELD CONTACTS", "Consensual contacts between officers and the community.")
    # Defintion from https://www.urban.org/sites/default/files/publication/25781/412647-Key-Issues-in-the-Police-Use-of-Pedestrian-Stops-and-Searches.PDF
    PEDESTRIAN = ("PEDESTRIAN STOPS",'"A street stop by an officer whereby an officer stops and questions a ' +
        "pedestrian based on reasonable suspicion that the pedestrian is or was recently engaged in unlawful " + 
        "activity. Street stops may lead to a limited search, called a “pat down” or frisk. If the officer " +
        "obtains probable cause that the pedestrian is or was involved in a crime, the stop could lead to a " +
        'full body search." See preface of https://www.urban.org/sites/default/files/publication/25781/412647-Key-Issues-in-the-Police-Use-of-Pedestrian-Stops-and-Searches.PDF')
    PEDESTRIAN_ARRESTS = ("PEDESTRIAN ARRESTS","Pedestrian stops table only containing data for incidents ending in arrests. " +
        "See TableType.PEDESTRIAN.description for the definitinon of a pedestrian stop.")
    PEDESTRIAN_CITATIONS = ("PEDESTRIAN CITATIONS","Pedestrian stops table only containing data for incidents ending in citations. " +
        "See TableType.CITATIONS.description for a general definition of a citation. See TableType.PEDESTRIAN.description " + 
        "for the definitinon of a pedestrian stop.")
    PEDESTRIAN_WARNINGS = ("PEDESTRIAN WARNINGS","Pedestrian stops table only containing data for incidents ending in warnings. " + 
        "A pedestrian warning is a verbal or written warning issued by a police " + 
        "officer durian a pedestrian stop. See TableType.PEDESTRIAN.description " + 
        "for the definitinon of a pedestrian stop.")
    # Definition from https://www.policedatainitiative.org/datasets/officer-involved-shootings/
    SHOOTINGS = ("OFFICER-INVOLVED SHOOTINGS",'"Although no national or standard definition exists, an officer ' + 
        "involved shooting (OIS) may be defined as the discharge of a firearm, which may include accidental and " +
        "intentional discharges, by a police officer, whether on or off duty. In some cases OIS datasets only " +
        "include instances in which an officer discharged a firearm at a person and may not include discharges " +
        'directed into or at a vehicle, animal, etc." https://www.policedatainitiative.org/datasets/officer-involved-shootings/')
    SHOOTINGS_CIVILIANS = ("OFFICER-INVOLVED SHOOTINGS - CIVILIANS",
        "Since multiple civilians and officers can be involved in a use of force incident, some departments " +
        "have separate tables for the incident, the civilians involved in the shooting, and the officers " + 
        "involved in the shooting. They are typically linked by a unique identifier. This table is for the " + 
        "civilians. See TableType.SHOOTINGS.description for definition of an officer-involved shooting.")
    SHOOTINGS_OFFICERS = ("OFFICER-INVOLVED SHOOTINGS - OFFICERS",
        "Since multiple civilians and officers can be involved in a use of force incident, some departments " +
        "have separate tables for the incident, the civilians involved in the shooting, and the officers " + 
        "involved in the shooting. They are typically linked by a unique identifier. This table is for the " + 
        "officers. See TableType.SHOOTINGS.description for definition of an officer-involved shooting.")
    SHOOTINGS_INCIDENTS = ("OFFICER-INVOLVED SHOOTINGS - INCIDENTS",
        "Since multiple civilians and officers can be involved in an officer-involved shooting, some departments " +
        "have separate tables for the incident, the civilians involved in the shooting, and the officers " + 
        "involved in the shooting. They are typically linked by a unique identifier. This table is for the " + 
        "incident. See TableType.SHOOTINGS.description for definition of an officer-involved shooting.")
    STOPS = ("STOPS","Table containing both pedestrian and traffic stops by officers. See TableType.TRAFFIC.description " +
        " and TableType.PEDESTRIAN.description for defintions of traffic and pedestrian stops, respectively.")
    TRAFFIC = ("TRAFFIC STOPS","Traffic stops are stops by police of motor vehicles due to reasonable suspicion " + 
        " or traffic violations.")
    TRAFFIC_ARRESTS = ("TRAFFIC ARRESTS","Traffic stops table for traffic stops ending in arrests. " +
        "See TableType.TRAFFIC.description for the definitinon of a traffic stop.")
    TRAFFIC_CITATIONS = ("TRAFFIC CITATIONS","Traffic stops table for traffic stops ending in citations. " +
        "See TableType.CITATIONS.description for a general definition of a citation and " +
        "TableType.TRAFFIC.description for the definitinon of a traffic stop.")
    # Defintion from https://legalbeagle.com/7700165-traffic-citation-warning.html
    TRAFFIC_WARNINGS = ("TRAFFIC WARNINGS","Traffic stops table for traffic stops ending in citations. " +
        '"A traffic citation warning is a verbal or written warning issued by a police ' + 
        'officer in the event of a traffic violation" (https://legalbeagle.com/7700165-traffic-citation-warning.html)' + 
        "See TableType.TRAFFIC.description for the definitinon of a traffic stop.")
    # Definition from https://www.policedatainitiative.org/use-of-force/
    USE_OF_FORCE = ("USE OF FORCE",'"The use of force can generally be defined as the means of compelling compliance or ' + 
        "overcoming resistance to an officer’s command(s) in order to protect life or property or to take a person into custody. " + 
        "For this reason, some agencies refer to the use of force as “response to resistance.” Types of force can include verbal, " + 
        'physical, chemical, impact, electronic, and firearm. Other definitions of the use of force in law enforcement may differ." https://www.policedatainitiative.org/use-of-force/')
    USE_OF_FORCE_CIVILIANS = ("USE OF FORCE - CIVILIANS",
        "Since multiple civilians and officers can be involved in a use of force incident, some departments " +
        "have separate tables for the use of force incident, the civilians involved in the incident, and the officers " + 
        "involved in the incident. They are typically linked by a unique identifier. This table is for the " + 
        "civilians. See TableType.USE_OF_FORCE.description for definition of use of force.")
    USE_OF_FORCE_OFFICERS = ("USE OF FORCE - OFFICERS",
        "Since multiple civilians and officers can be involved in a use of force incident, some departments " +
        "have separate tables for the use of force incident, the civilians involved in the incident, and the officers " + 
        "involved in the incident. They are typically linked by a unique identifier. This table is for the " + 
        "officers. See TableType.USE_OF_FORCE.description for definition of use of force.")
    USE_OF_FORCE_INCIDENTS = ("USE OF FORCE - INCIDENTS",
        "Since multiple civilians and officers can be involved in a use of force incident, some departments " +
        "have separate tables for the use of force incident, the civilians involved in the incident, and the officers " + 
        "involved in the incident. They are typically linked by a unique identifier. This table is for the " + 
        "incident. See TableType.USE_OF_FORCE.description for definition of use of force.")
    USE_OF_FORCE_CIVILIANS_OFFICERS = ("USE OF FORCE - CIVILIANS/OFFICERS",
        "Since multiple civilians and officers can be involved in a use of force incident, some departments " +
        "have separate tables for the use of force incident, the civilians involved in the incident, and the officers " + 
        "involved in the incident. They are typically linked by a unique identifier. This table contains data for both the " + 
        "civilians and the officers. See TableType.USE_OF_FORCE.description for definition of use of force.")
    # Definition from https://www.police.ucla.edu/other/vehicle-pursuits
    VEHICLE_PURSUITS = ("VEHICLE PURSUITS",'"A vehicle pursuit is an event involving one or more law enforcement officers ' +
        "attempting to apprehend a suspect who is attempting to avoid arrest while operating a motor vehicle by using high " +
        "speed driving or other evasive tactics such as driving off a highway, turning suddenly or driving in a legal manner " +
        'but willfully failing to yield to an officer’s signal to stop" https://www.police.ucla.edu/other/vehicle-pursuits')

# Constants used in dataset parameters
MULTI = "MULTI"    # For data sets that put multiple years or agencies in 1 dataset
NA = "NONE"         # None = not applicable (pandas converts "N/A" to NaN)

_col_names = [
    "DATE", 
    "CIVILIAN_RACE",
    "CIVILIAN_ETHNICITY",
    "OFFICER_RACE",
    "OFFICER_ETHNICITY",
    "AGENCY"
]
columns = namedtuple('Columns',
    _col_names,
    defaults=_col_names
    )()

_race_names = [
    "AAPI",
    "ASIAN",
    "ASIAN_INDIAN",
    "BLACK",
    "HAWAIIAN",
    "LATINO",
    "MIDDLE_EASTERN",
    "MULTIPLE",
    "NATIVE_AMERICAN",
    "OTHER",
    "OTHER_UNKNOWN",
    "UNKNOWN",
    "UNSPECIFIED",
    "WHITE"
]
_map = {
    "AAPI" : "ASIAN / PACIFIC ISLANDER",
    "ASIAN_INDIAN" : "ASIAN INDIAN",
    "HAWAIIAN" : "HAWAIIAN / PACIFIC ISLANDER",
    "LATINO" : "HISPANIC / LATINO",
    "MIDDLE_EASTERN" : "MIDDLE EASTERN",
    "NATIVE_AMERICAN" : "NATIVE AMERICAN",
    "OTHER_UNKNOWN" : "OTHER / UNKNOWN"
}
_race_defaults = [_map.get(x,x) for x in _race_names]
races = namedtuple('Races',
    _race_names,
    defaults=_race_defaults
    )()


import os
from datetime import date
import pandas as pd
from numpy import nan
import requests
import urllib
from sodapy import Socrata
from pyproj.exceptions import CRSError
from pyproj import CRS
import warnings
from arcgis.features import FeatureLayerCollection
from arcgis.geometry._types import Point
from time import sleep
from tqdm import tqdm
from math import ceil

try:
    import geopandas as gpd
    _has_gpd = True
except:
    _has_gpd = False

try:
    from .exceptions import OPD_TooManyRequestsError, OPD_DataUnavailableError, OPD_arcgisAuthInfoError, OPD_SocrataHTTPError
except:
    from exceptions import OPD_TooManyRequestsError, OPD_DataUnavailableError, OPD_arcgisAuthInfoError, OPD_SocrataHTTPError

sleep_time = 0.1

# Global parameter for testing both with and without GeoPandas in testing
_use_gpd_force = None

_default_limit = 100000
_url_error_msg = "There is likely an issue with the website. Open the URL {} with a web browser to confirm. " + \
                    "See a list of known site outages at https://github.com/openpolicedata/opd-data/blob/main/outages.csv"

# This is for use if import data sets using Socrata. It is not required.
# Requests made without an app_token will be subject to strict throttling limits
# Get a App Token here: http://dev.socrata.com/docs/app-tokens.html
# Copy the App Token
# Create an environment variable SODAPY_API_KEY and set it equal to the API key
# Setting environment variable in Linux: https://phoenixnap.com/kb/linux-set-environment-variable
# Windows: https://www.wikihow.com/Create-an-Environment-Variable-in-Windows-10
default_sodapy_key = os.environ.get("SODAPY_API_KEY")

def load_csv(url, date_field=None, year_filter=None, agency_field=None, agency=None, limit=None, pbar=True):
    '''Download CSV file to pandas DataFrame
    
    Parameters
    ----------
    url : str
        Download URL for CSV
    date_field : str
        (Optional) Name of the column that contains the date
    year_filter : int, list
        (Optional) Either the year or the year range [first_year, last_year] for the data that is being requested. None value returns data for all years.
    agency_field : str
        (Optional) Name of the column that contains the agency name (i.e. name of the police departments)
    agency : str
        (Optional) Name of the agency to filter for. None value returns data for all agencies.
    limit : int
        (Optional) Only returns the first limit rows of the CSV
    pbar : bool
        (Optional) If true (default), a progress bar will be displayed
        
    Returns
    -------
    pandas DataFrame
        DataFrame containing table imported from CSV
    '''
    
    if ".zip" in url or not pbar:
        with warnings.catch_warnings():
            # Perhaps use requests iter_content/iter_lines as below to read large CSVs so progress can be shown
            warnings.simplefilter("ignore", category=pd.errors.DtypeWarning)
            try:
                table = pd.read_csv(url, encoding_errors='surrogateescape')
            except urllib.error.HTTPError as e:
                raise OPD_DataUnavailableError(*e.args, _url_error_msg.format(url))
            except Exception as e:
                raise e
    else:
        r = requests.head(url)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise OPD_DataUnavailableError(*e.args, _url_error_msg.format(url))
        except Exception as e:
            raise e
        with requests.get(url, params=None, stream=True) as resp:
            try:
                table = pd.read_csv(TqdmReader(resp), nrows=limit, encoding_errors='surrogateescape')
            except Exception as e:
                raise e

    table = filter_dataframe(table, date_field=date_field, year_filter=year_filter, 
        agency_field=agency_field, agency=agency)

    return table


def load_arcgis(url, date_field=None, year=None, limit=None, pbar=True):
    '''Download table from ArcGIS to pandas or geopandas DataFrame
    
    Parameters
    ----------
    url : str
        ArcGIS GeoService URL
    date_field : str
        (Optional) Name of the column that contains the date
    year : int, list
        (Optional) Either the year or the year range [first_year, last_year] for the data that is being requested.  None value returns data for all years.
    limit : int
        (Optional) Only returns the first limit rows of the table
    pbar : bool
        (Optional) If true (default), a progress bar will be displayed
        
    Returns
    -------
    pandas or geopandas DataFrame
        DataFrame containing table imported from ArcGIS
    '''
    
    # Table vs. Layer: https://developers.arcgis.com/rest/services-reference/enterprise/layer-feature-service-.htm
    # The layer resource represents a single feature layer or a nonspatial table in a feature service. 
    # A feature layer is a table or view with at least one spatial column.
    # For tables, it provides basic information about the table such as its ID, name, fields, types, and templates. 
    # For feature layers, in addition to the table information, it provides information such as its geometry type, min and max scales, and spatial reference.

    if url[-1] == "/":
        url = url[0:-1]
    last_slash = url.rindex("/")
    layer_num = url[last_slash+1:]
    base_url = url[:last_slash]
    
    user_limit = limit != None
    if not user_limit:
        limit = _default_limit
    else:
        total = limit

    # Get metadata
    r = requests.get(base_url + "/" + layer_num + "?f=pjson")
    r.raise_for_status()
    meta = r.json()
    if "maxRecordCount" in meta and \
        (not user_limit or (user_limit and limit > meta["maxRecordCount"])):
        limit = meta["maxRecordCount"]
    
    # https://developers.arcgis.com/python/
    try:
        layer_collection = FeatureLayerCollection(base_url)
    except Exception as e:
        if len(e.args)>0:
            if "Error Code: 500" in e.args[0]:
                raise OPD_DataUnavailableError(base_url, f"Layer # = {layer_num}", e.args)
            elif "A general error occurred: 'authInfo'" in e.args[0]:
                raise OPD_arcgisAuthInfoError(base_url, f"Layer # = {layer_num}", e.args)
        else: raise e
    except e: raise e

    is_table = True
    active_layer = None
    for layer in layer_collection.layers:
        layer_url = layer.url
        if layer_url[-1] == "/":
            layer_url = layer_url[:-1]
        if layer_num == layer_url[last_slash+1:]:
            active_layer = layer
            is_table = False
            break

    if is_table:
        for layer in layer_collection.tables:
            layer_url = layer.url
            if layer_url[-1] == "/":
                layer_url = layer_url[:-1]
            if layer_num == layer_url[last_slash+1:]:
                active_layer = layer
                break

    if active_layer == None:
        raise ValueError("Unable to find layer")
    
    where_query = ""
    if date_field!=None and year!=None:
        start_date, stop_date = _process_date(year, inclusive=False)
        
        where_query = f"{date_field} >= '{start_date}' AND  {date_field} < '{stop_date}'"
    else:
        where_query = '1=1'

    try:
        record_count = active_layer.query(where=where_query, return_count_only=True)
        if record_count==0:
            return None
    except Exception as e:
        if len(e.args)>0 and "Error Code: 429" in e.args[0]:
            raise OPD_TooManyRequestsError(base_url, f"Layer # = {layer_num}", *e.args, _url_error_msg.format(url))
        else:
            raise
    except:
        raise

    if user_limit:
        num_batches = ceil(total / limit)
    else:
        num_batches = ceil(record_count / limit)
        total = record_count
        
    df = []
    pbar = pbar and num_batches>1
    if pbar:
        bar = tqdm(desc=url, total=total, leave=False) 
        
    for batch in range(num_batches):
        try:
            if batch==0:
                layer_query_result = active_layer.query(where=where_query, result_offset=batch*limit, result_record_count=limit, return_all_records=False)
                df.append(layer_query_result.sdf)
                if len(df[0]) not in [limit, total]:
                    num_rows = len(df[0])
                    raise ValueError(f"Number of rows is {num_rows} but is expected to be max rows to read {limit} or total number of rows {total}")
            else:
                df.append(active_layer.query(where=where_query, result_offset=batch*limit, result_record_count=limit, return_all_records=False, as_df=True))
        except Exception as e:
            if len(e.args)>0 and "Error Code: 429" in e.args[0]:
                raise OPD_TooManyRequestsError(base_url, f"Layer # = {layer_num}", *e.args, _url_error_msg.format(url))
            else:
                raise
        except:
            raise

        if pbar:
            bar.update(len(df[-1]))

    if pbar:
        bar.close()

    df = pd.concat(df, ignore_index=True)

    if len(df) > 0:
        if is_table:
            if "SHAPE" in df:
                raise ValueError("Tables are not expected to include geographic data")
            return df
        else:
            if _use_gpd_force is not None:
                if not _has_gpd and _use_gpd_force:
                    raise ValueError("User cannot force GeoPandas usage when it is not installed")
                use_gpd = _use_gpd_force
            else:
                use_gpd = _has_gpd

            if use_gpd:
                def fix_nans(pt):
                    if type(pt) == Point and pt.x=="NaN":
                        pt.x = nan
                        pt.y = nan

                    return pt
                geometry = df.pop("SHAPE").apply(fix_nans)
                try:
                    df = gpd.GeoDataFrame(df, crs=layer_query_result.spatial_reference['wkid'], geometry=geometry)
                except CRSError:
                    # Method recommended by pyproj to deal with CRSError for wkid = 102685
                    crs = CRS.from_authority("ESRI", layer_query_result.spatial_reference['wkid'])
                    df = gpd.GeoDataFrame(df, crs=crs, geometry=geometry)

            # try:
            #     json_data = layer_query_result.to_geojson
            # except:
            #     for k in range(len(layer_query_result.features)):
            #         if layer_query_result.features[k].geometry == None:
            #             # Put in dummy data
            #             layer_query_result.features[k].geometry = {"x" : nan, "y" : nan}

            #     json_data = layer_query_result.to_geojson

            # json_data = json.loads(json_data)
            
            # for k in range(len(json_data["features"])):
            #     if json_data["features"][k]["geometry"]['coordinates'] == ["NaN", "NaN"]:
            #         json_data["features"][k]["geometry"]['coordinates'] = [nan, nan]
            
            # if _use_gpd_force is not None:
            #     if not _has_gpd and _use_gpd_force:
            #         raise ValueError("User cannot force GeoPandas usage when it is not installed")
            #     use_gpd = _use_gpd_force
            # else:
            #     use_gpd = _has_gpd
                
            # if use_gpd:
            #     try:
            #         df = gpd.GeoDataFrame.from_features(json_data, crs=layer_query_result.spatial_reference['wkid'])
            #     except CRSError:
            #         # Method recommended by pyproj to deal with CRSError for wkid = 102685
            #         crs = CRS.from_authority("ESRI", layer_query_result.spatial_reference['wkid'])
            #         df = gpd.GeoDataFrame.from_features(json_data, crs=crs)
            # else:
            #     dict_data = [x['properties'] for x in json_data['features']]
            #     for k in range(len(dict_data)):
            #         if 'geometry' in json_data['features'][k]:
            #             dict_data[k]['geometry'] = json_data['features'][k]['geometry']

            #     df = pd.DataFrame.from_records(dict_data)

            # This should be kept?
            # if date_field is not None:
            #     date_field_metadata=[x for x in layer_query_result.fields if x['name']==date_field]
            #     if len(date_field_metadata) != 1:
            #         raise ValueError(f"Unable to find a single date field named {date_field}. Found {len(date_field_metadata)} instances.")

            #     if date_field_metadata[0]['type'] in ['esriFieldTypeDate', "esriFieldTypeString"]:
            #         df = df.astype({date_field: 'datetime64[ms]'})
            #     else:
            #         raise ValueError(f"Unsupported data type {date_field_metadata[0]['type']} for field {date_field}.")
            

            return df
    else:
        return None


def load_socrata(url, data_set, date_field=None, year=None, opt_filter=None, select=None, output_type=None, 
                 limit=None, key=default_sodapy_key, pbar=True):
    '''Download table from Socrata to pandas or geopandas DataFrame
    
    Parameters
    ----------
    url : str
        URL for Socrata data
    data_set : str
        Dataset ID for Socrata data
    date_field : str
        (Optional) Name of the column that contains the date
    year : int, list
        (Optional) Either the year or the year range [first_year, last_year] for the data that is being requested.  None value returns data for all years.
    opt_filter : str
        (Optional) Additional filter to apply to data (beyond any date filter specified by date_field and year)
    select : str
        (Optional) select statement to REST API
    output_type : str
        (Optional) Data type for the output. Default: pandas or geopandas DataFrame
    limit : int
        (Optional) Only returns the first limit rows of the table
    key : str
        (Optional) Socrata app token to prevent throttling of the data request
    pbar : bool
        (Optional) If true (default), a progress bar will be displayed
        
    Returns
    -------
    pandas or geopandas DataFrame
        DataFrame containing table
    '''

    # Unauthenticated client only works with public data sets. Note 'None'
    # in place of application token, and no username or password:
    client = Socrata(url, key, timeout=60)

    user_limit = limit != None
    if not user_limit:
        limit = _default_limit

    N = 1  # Initialize to value > 0 so while loop runs
    offset = 0

    where = ""
    if date_field!=None and year!=None:
        start_date, stop_date = _process_date(year,inclusive=True, date_field=date_field)
        where = date_field + " between '" + start_date + "' and '" + stop_date +"'"

    if opt_filter is not None:
        if not isinstance(opt_filter, list):
            opt_filter = [opt_filter]

        andStr = " AND "
        for filt in opt_filter:
            where += andStr + filt

        if where[0:len(andStr)] == andStr:
            where = where[len(andStr):]

    if _use_gpd_force is not None:
        if not _has_gpd and _use_gpd_force:
            raise ValueError("User cannot force GeoPandas usage when it is not installed")
        use_gpd = _use_gpd_force
    else:
        use_gpd = _has_gpd
         
    show_pbar = pbar and not user_limit and select==None
    if show_pbar:
        results = client.get(data_set, where=where, select="count(*)")
        try:
            num_rows = float(results[0]["count"])
        except:
            num_rows = float(results[0]["count_1"]) # Value used in VT Shootings data
        total = ceil(num_rows / limit)
        if total > 1:
            bar = tqdm(desc=f"URL: {url}, Dataset: {data_set}", total=total, leave=False)
        else:
            show_pbar = False

    order = None
    if select == None:
        # order guarantees data order remains the same when paging
        # Order by date if available otherwise the data ID. 
        # https://dev.socrata.com/docs/paging.html#2.1
        order = ":id" if date_field==None else date_field

    while N > 0:
        try:
            results = client.get(data_set, where=where,
                limit=limit,offset=offset, select=select, order=order)
        except requests.HTTPError as e:
            raise OPD_SocrataHTTPError(url, data_set, *e.args, _url_error_msg.format(url))
        except Exception as e: 
            if len(e.args)>0 and e.args[0]=='Unknown response format: text/html':
                raise OPD_SocrataHTTPError(url, data_set, *e.args, _url_error_msg.format(url))
            else:
                raise e

        if use_gpd and output_type==None:
            # Check for geo info
            for r in results:
                if "geolocation" in r or "geocoded_column" in r:
                    output_type = "GeoDataFrame"
                    break

        if output_type == "set":
            if offset==0:
                df = set()

            if len(results)>0:
                filt_key = select.replace("DISTINCT ", "")
                results = [row[filt_key] for row in results if len(row)>0]
                results = set(results)
                df.update(results)

        elif output_type == "list":
            if offset==0:
                df = list()

            if len(results)>0:
                [df.append(row[select]) for row in results]

        elif use_gpd and output_type=="GeoDataFrame":
            output_type = "GeoDataFrame"
            # Presumed to be a list of properties that possibly include coordinates
            geojson = {"type" : "FeatureCollection", "features" : []}
            for p in results:
                feature = {"type" : "Feature", "properties" : p}
                if "geolocation" in feature["properties"]:
                    geo = feature["properties"].pop("geolocation")
                    if list(geo.keys()) == ["human_address"]:
                        feature["geometry"] = {"type" : "Point", "coordinates" : (nan, nan)}  
                    elif "coordinates" in geo:
                        feature["geometry"] = geo
                    else:
                        feature["geometry"] = {"type" : "Point", "coordinates" : (float(geo["longitude"]), float(geo["latitude"]))}
                elif "geocoded_column" in feature["properties"]:
                    feature["geometry"] = feature["properties"].pop("geocoded_column")
                else:
                    feature["geometry"] = {"type" : "Point", "coordinates" : (nan, nan)} 
                
                geojson["features"].append(feature)

            if len(results)>0:
                new_gdf = gpd.GeoDataFrame.from_features(geojson, crs=4326)
                    
                if offset==0:
                    df = new_gdf
                else:
                    df = pd.concat([df, new_gdf], ignore_index=True)
        else:
            output_type = "DataFrame"
            rows = pd.DataFrame.from_records(results)
            if offset==0:
                df = pd.DataFrame(rows)
            else:
                df = pd.concat([df, rows], ignore_index=True)

        N = len(results)
        offset += N

        if user_limit:
            break
        if show_pbar:
            bar.update()

    if show_pbar:
        bar.close()
    return df


def _process_date(date, inclusive, date_field=None):
    if not isinstance(date, list):
        date = [date, date]

    if len(date) !=2:
        raise ValueError("date should be a list of length 2: [startYear, stopYear]")

    if date[0] > date[1]:
        raise ValueError('date[0] needs to be smaller than or equal to date[1]')

    if type(date[0]) == str:
        # This should already be in date format
        start_date = date[0]
    elif date_field != None and date_field.lower() == "year":
        # Assuming this as actually a string or numeric field for the year rather than a datestamp
        start_date = str(date[0])
    else:
        start_date = str(date[0]) + "-01-01"

    if type(date[1]) == str:
        # This should already be in date format
        stop_date = date[1]
    elif date_field != None and date_field.lower() == "year":
        # Assuming this as actually a string or numeric field for the year rather than a datestamp
        stop_date = str(date[1])
    else:
            stop_date  = str(date[1]) + "-12-31T23:59:59.999"

    return start_date, stop_date


def filter_dataframe(df, date_field=None, year_filter=None, agency_field=None, agency=None):
    '''Load CSV file to pandas DataFrame
    
    Parameters
    ----------
    df : pandas or geopandas dataframe
        Dataframe containing the data
    date_field : str
        (Optional) Name of the column that contains the date
    year_filter : int, list
        (Optional) Either the year or the year range [first_year, last_year] for the data that is being requested.  None value returns data for all years.
    agency_field : str
        (Optional) Name of the column that contains the agency name (i.e. name of the police departments)
    agency : str
        (Optional) Name of the agency to filter for. None value returns data for all agencies.
    '''
    
    if year_filter != None and date_field != None:
        df = df[df[date_field].dt.year == year_filter]

    if agency != None and agency_field != None:
        df = df.query(agency_field + " = '" + agency + "'")

    return df


def get_years_argis(url, date_field):
    '''Returns the years of the data contained in an ArcGIS table
    
    Parameters
    ----------
    url : str
        ArcGIS GeoService URL
    date_field : str
        Name of the column that contains the date
        
    Returns
    -------
    list
        List of years contained in the table
    '''
    
    return _get_years("arcgis", url, date_field=date_field)


def get_years_socrata(url, data_set, date_field):
    '''Returns the years of the data contained in a Socrata table
    
    Parameters
    ----------
    url : str
        URL of Socrata table
    date_field : str
        Name of the column that contains the date
        
    Returns
    -------
    list
        List of years contained in the table
    '''
    
    return _get_years("socrata", url, date_field=date_field, data_set=data_set)


def _get_years(data_type, url, date_field, data_set=None):
    year = date.today().year
    data_type = data_type.lower()

    oldest_recent = 20
    max_misses_gap = 10
    max_misses = oldest_recent
    misses = 0
    years = []
    while misses < max_misses:
        if data_type == "arcgis":
            df = load_arcgis(url, date_field=date_field, year=year, limit=1)
        elif data_type == "socrata":
            df = load_socrata(url, data_set, date_field=date_field, year=year, limit=1)
        else:
            raise ValueError("Unknown data type")

        if not hasattr(df, '__len__') or len(df)==0:  # If doesn't have len attribute, it is None
            misses+=1
        else:
            misses = 0
            max_misses = max_misses_gap
            years.append(year)

        sleep(sleep_time)

        year-=1

    return years

# https://stackoverflow.com/questions/73093656/progress-in-bytes-when-reading-csv-from-url-with-pandas
class TqdmReader:
    def __init__(self, resp, limit=None):
        total_size = int(resp.headers.get("Content-Length", 0))

        self.rows_read = 0
        if limit != None:
            self.limit = limit
        else:
            self.limit = float("inf")
        self.resp = resp
        self.bar = tqdm(
            desc=resp.url,
            total=total_size,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
            leave=False
        )

        self.reader = self.read_from_stream()

    def read_from_stream(self):
        for line in self.resp.iter_lines():
            line += b"\n"
            self.bar.update(len(line))
            yield line

    def read(self, n=0):
        try:
            if self.rows_read >= self.limit:
                # Number of rows read is greater than user-requested limit
                return ""
            self.rows_read += 1
            return next(self.reader)
        except StopIteration:
            self.bar.update(self.bar.total - self.bar.n)
            return ""

if __name__ == "__main__":
    import time
    _default_limit = 10000
    start_time = time.time()
    url = "https://gis.charlottenc.gov/arcgis/rest/services/CMPD/CMPD/MapServer/16/"
    url = 'https://services1.arcgis.com/zdB7qR0BtYrg0Xpl/arcgis/rest/services/ODC_CRIME_STOPS_P/FeatureServer/32'
    date_field = 'TIME_PHONEPICKUP'
    load_arcgis(url,date_field=date_field, year=2020)
    print(f"Completed in {time.time()-start_time} seconds")



import pandas as pd
import numpy as np
import re
import warnings

try:
    from .defs import TableType, DataType
except:
    from defs import TableType, DataType

# Location of table where datasets available in opd are stored
csv_file = "https://raw.github.com/openpolicedata/opd-data/main/opd_source_table.csv"

def _build(csv_file):
    # Check columns
    columns = {
        'State' : pd.StringDtype(),
        'SourceName' : pd.StringDtype(),
        'Agency': pd.StringDtype(),
        'TableType': pd.StringDtype(),
        'Year': np.dtype("O"),
        'Description': pd.StringDtype(),
        'DataType': pd.StringDtype(),
        'URL': pd.StringDtype(),
        'date_field': pd.StringDtype(),
        'dataset_id': pd.StringDtype(),
        'agency_field': pd.StringDtype(),
        'readme': pd.StringDtype(),
        'min_version': pd.StringDtype()
    }

    try:
        df = pd.read_csv(csv_file, dtype=columns)
    except:
        warnings.warn(f"Unable to load CSV file from {csv_file}. " +
            "This may be due to a bad internet connection or bad filename/URL.")
        return None

    if "Jurisdiction" in df:
        df.rename(columns={
            "Jurisdiction" : "Agency",
            "jurisdiction_field" : "agency_field"
        }, inplace=True)

    # Convert years to int
    df["Year"] = [int(x) if x.isdigit() else x for x in df["Year"]]
    df["SourceName"] = df["SourceName"].str.replace("Police Department", "")
    df["Agency"] = df["Agency"].str.replace("Police Department", "")

    for col in df.columns:
        df[col] = [x.strip() if type(x)==str else x for x in df[col]]

    # ArcGIS datasets should have a URL ending in either /FeatureServer/# or /MapServer/#
    # Where # is a layer #
    urls = df["URL"]
    p = re.compile(r"(MapServer|FeatureServer)/\d+")
    for i,url in enumerate(urls):
        if df.iloc[i]["DataType"] == DataType.ArcGIS.value:
            result = p.search(url)
            urls[i] = url[:result.span()[1]]

    df["URL"] = urls

    keyVals = ['State', 'SourceName', 'Agency', 'TableType','Year']
    df.drop_duplicates(subset=keyVals, inplace=True)
    # df.sort_values(by=keyVals, inplace=True, ignore_index=True)

    return df


datasets = _build(csv_file)


# Datasets that had issues that need added in the future
# _builder.add_data(state="North Carolina", agency="Charlotte-Mecklenburg",
#     table_type=TableType.TRAFFIC, 
#     url=["https://gis.charlottenc.gov/arcgis/rest/services/CMPD/CMPD/MapServer/14/"], 
#     data_type=DataType.ArcGIS,
#     description="Traffic Stops",
#     lut_dict={"date_field" : "Month_of_Stop"})
# _builder.add_data(state="Vermont", agency="Burlington",
#     tableType=TableType.USE_OF_FORCE, 
#     url=["https://data.burlingtonvt.gov/explore/dataset/bpd-use-of-force/"], 
#     data_type=DataType.UNKNOWN,
#     description="Use-of-Force incidents",
#     lut_dict={"date_field" : "call_time"})
# _builder.add_data(state="Vermont", agency="Burlington",
#     tableType=TableType.TRAFFIC, 
#     url=["https://data.burlingtonvt.gov/explore/dataset/bpd-traffic-stops/"], 
#     data_type=DataType.UNKNOWN,
#     description="Traffic Stops",
#     lut_dict={"date_field" : "call_time"})
# _builder.add_data(state="Vermont", agency="Burlington",
#     tableType=TableType.ARRESTS, 
#     url=["https://data.burlingtonvt.gov/explore/dataset/arrests/"], 
#     data_type=DataType.UNKNOWN,
#     description="Arrests",
#     lut_dict={"date_field" : "arrest_date"})
# _builder.add_data(state="Vermont", agency="Burlington",
#     tableType=TableType.ARRAIGNMENT, 
#     url=["https://data.burlingtonvt.gov/explore/dataset/arraignment-and-bail-data/"], 
#     data_type=DataType.UNKNOWN,
#     description="Case level data set on arraignment and bail",
#     lut_dict={"date_field" : "arraignment_date"})
# _builder.add_data(state="California", source_name="California Department of Justice", agency=MULTI,
#     tableType=TableType.DEATHES_IN_CUSTODY, 
#     url=["https://data-openjustice.doj.ca.gov/sites/default/files/dataset/2021-07/DeathInCustody_2005-2020_20210603.xlsx"], 
#     data_type=DataType.EXCEL,
#     escription="State and local law enforcement agencies and correctional facilities report information on deaths that occur in custody or during the process of arrest in compliance with Section 12525 of the California Government Code",
#     lut_dict={"date_field" : "date_of_death_yyyy"})


def datasets_query(source_name=None, state=None, agency=None, table_type=None):
    """Query for available datasets.
    Request a DataFrame containing available datasets based on input filters.
    Returns all datasets if no filters applied.
    
    Parameters
    ----------
    source_name : str
        OPTIONAL name of source to filter by source name
    state : str
        OPTIONAL name of state to filter by state
    agency : str
        OPTIONAL name of agency to filter by agency
    table_type : str or TableType enum
        OPTIONAL name of table type to filter by type of data

    RETURNS
    -------
    Dataframe containing datasets that match any filters applied
    """
    query = ""
    if state != None:
        query += "State == '" + state + "' and "

    if source_name != None:
        query += "SourceName == '" + source_name + "' and "

    if agency != None:
        query += "Agency == '" + agency + "' and " 

    if table_type != None:
        if isinstance(table_type, TableType):
            table_type = table_type.value
        query += "TableType == '" + table_type + "' and "

    if len(query) == 0:
        return datasets.copy()
    else:
        return datasets.query(query[0:-5]) 


if __name__=="__main__":
    df = datasets_query()
    df = datasets_query("Virginia")
